/**CCSA.cpp***********************************************************************************************
 *          Copyright 2004, V. Makinen                                                                   *
 *                                                                                                       *
 * This file contains implementations of two succinct full-text indexes: CCSA and FMindex.               *
 * For more information, see [MN04a,MN04b], respectively                                                 *                     
 *                                                                                                       *
 *                                                                                                       *   
 * This software may be used freely for any purpose. However, when distributed,                          *  
 * the original source must be clearly stated, and, when the source code is                              *
 * distributed, the copyright notice must be retained and any alterations in                             *
 * the code must be clearly marked. No warranty is given regarding the quality                           *         
 * of this software.                                                                                     *
 *                                                                                                       *
 * [MN04a] Veli Mkinen and Gonzalo Navarro. Compressed Compact Suffix Arrays. Submitted, February 2004. *
 * [MN04b] Veli Mkinen and Gonzalo Navarro. On Succinct Suffix Arrays. Submitted, February 2004.        *
 *                                                                                                       *
 *********************************************************************************************************/


//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>
#include <assert.h>
#include <string.h>
#include "qsufsort.c"
//---------------------------------------------------------------------------

#define W 32 // computer word size
#define Wminusone 31  // W-1
#define WW 64  // 2*W

unsigned n,m; // text length, pattern length
bool print=true; 
int report=1;  // 0=existence, 1=report, 2=count, X>2 = print X characters after each pattern occurrence
unsigned bitsn=0; // bits needed to represent n

char* text; 

class TSA;
class TBitRank;
class TAlphabetRank;
class TCCSA;
class TFMindex;
char *BWT(char *text, unsigned n);
void TestAlphabetRank();
TSA *CreateSuffixArray(char *text, unsigned n);
TCCSA *MakeSuffixArrayAndCompressIt(char *text, unsigned n);

/**********************************************************************************************************
 * General code, common to CCSA and to FMindex                                                            *
 **********************************************************************************************************/

inline void SetField(unsigned *A,register unsigned len, register unsigned index,register  unsigned x) {
   unsigned i=index*len/W, 
            j=index*len-i*W;
   unsigned mask = (j+len < W ? ~0u << j+len : 0) 
                   | (W-j < W ? ~0u >> (W-j) : 0);
   A[i] = (A[i] & mask) | x << j;
   if (j+len>W) {
      mask = ((~0u) << (len+j-W));  
      A[i+1] = (A[i+1] & mask)| x >> (W-j);
   } 
}; 

inline unsigned GetField(unsigned *A, register  unsigned len, register unsigned index) {
   register unsigned i=index*len/W, 
            j=index*len-W*i,
            result;                
   if (j+len <= W)
      result = (A[i] << (W-j-len)) >> (W-len);
   else {
      result = A[i] >> j;
      result = result | (A[i+1] << (WW-j-len)) >> (W-len);
   }
   return result;
};

inline unsigned bitsneeded(unsigned x) {return (unsigned)ceil(log((double)x)/log(2.0))+1;};

class TSA {
public:
    unsigned posbits;
    unsigned *pos;
    unsigned *B;    
    unsigned n;
public:
    void CreatePosArray(unsigned bits, unsigned n) 
      {this->n = n; posbits = bits; pos = new unsigned[n*posbits/W+2];
       CreateBooleanArray(n);};
    void CreateBooleanArray(unsigned n)
      {B = new unsigned[n/W+2];};   
    void DeletePosArray() {delete pos;};
    void DeleteBooleanArray() {delete B;};        
    void SetPos(unsigned index, unsigned x)
      {SetField(pos,posbits,index,x);};
    void SetB(unsigned index, unsigned x)
      {SetField(B,1,index,x);};
    unsigned GetPos(unsigned index)
      {return GetField(pos,posbits,index);};        
    bool GetB(unsigned index)
      {return GetField(B,1,index);};        
    ~TSA() {DeletePosArray(); DeleteBooleanArray();};
};

TSA *CreateSuffixArray(char *text, unsigned n)
{
   TSA *SA;
   int *x, *p;
   p=(int*)malloc((n+1)*sizeof *p);
   x=(int*)malloc((n+1)*sizeof *x);
   unsigned i;
   for (i=0; i<n; i++)
      x[i]=(int)text[i]+128;
   x[n]=0;
   suffixsort(x,p,n,256,0);
   free(x);
   unsigned k=0;
   while ((n<<k)>>(W-1) == 0) k++;
   bitsn = W-k;
   SA = new TSA();
   SA->CreatePosArray(bitsn,n);
   for (i=0; i<n; i++) {
      SA->SetPos(i,p[i+1]);
      SA->SetB(i,0);
   }
   free(p);
   return SA;
}


class TBitRank {
/* Supports rank(i)=(sum_{0<=i'<=i} B[i]) operation in constant time on bit-array B */
/* Takes less than n/(log n)+2n(loglog n)/(log n)+sqrt(n)(log n)(loglog n)+n bits */
/* That is, it takes n+o(n) bits (in practice about 1.5n bits) */   
/* Also supports operation select(x)=i, where rank(i)=x. */
/* Select is implemented making a binary search over the rank-structure, 
/* so it takes O(log(n)) time. It is possible to implement O(1)-time select,
/* using similar techniques as with rank.*/
/* See more details in David Clark's PhD thesis. */
public:   
   bool *B; // bit-array
   unsigned n; // length of B
   unsigned *firstrank; // rank of each log^2 n:th bit
   unsigned firstbits; // field length of firstrank (log n)
   unsigned firstn; // length of firstrank (n/log^2 n)/W
   unsigned *secondrank; // rank of each log n:th bit inside log^2 n size field
   unsigned secondbits; // field length of firstrank (2 log log n)
   unsigned secondn; // length of secondrank (2 n log log n/log n)/W
   unsigned *thirdrank; // ranks of all sqrt(n) fields of length (log n)/2  
   unsigned thirdbits;  // field length of thirdrank log((log n)/2)
   unsigned thirdn;  // length of thirdrank sqrt(n)(log((log n)/2)(log n)/2)/W
   unsigned *data; // stored version of B
   unsigned databits;
public:   
   TBitRank(bool *B, unsigned n) {this->B=B; this->n=n;};
   unsigned SpaceRequirementInBits() {
      return firstn*W+secondn*W+thirdn*W+n;
   }; 
   void MakeRankDictionary() {
      unsigned i,j;
      unsigned sum=0;
      /* First-level dictionary */  
      firstbits = bitsneeded(n);
      firstn = (unsigned)floor((floor(n/(1.0*firstbits))+1)*1.0/W+1)+1;
      firstrank = new unsigned[firstn];      
      j=0;
      for (i=0;i*firstbits*firstbits<n+firstbits*firstbits;i++) {
         for (;j<=i*firstbits*firstbits && j<n;j++)
            sum += B[j];
         SetField(firstrank,firstbits,i,sum); 
      }    
      /* Second-level dictionary */
      secondbits = bitsneeded(firstbits*firstbits);
      secondn = (unsigned)floor((floor(secondbits*n/(1.0*firstbits))+1)*1.0/W+1)+1+firstbits*secondbits/W+1;
      secondrank = new unsigned[secondn];
      sum = 0;
      j = 0;
      for (i=0;i*firstbits<n+firstbits*firstbits;i++) {
         if ((i*firstbits % (firstbits*firstbits))==0) {
            j = i*firstbits+1;
            sum = 0;
         }
         for (;j<=i*firstbits && j<n;j++)
            sum += B[j];
         SetField(secondrank,secondbits,i,sum); 
      }    
        /* Third-level dictionary */ 
      unsigned lognpertwo = (unsigned)floor(firstbits/2.0)+1; 
       unsigned x;
      thirdbits = bitsneeded(lognpertwo);
      unsigned thirdlength = (1u << lognpertwo); 
      thirdn = (unsigned)floor(1.0*thirdlength*lognpertwo*thirdbits/W)+1;
      thirdrank = new unsigned[thirdn];
       for (i=0;i<thirdlength;i++) {
          sum = 0;
          for (j=0;j<lognpertwo;j++) {
            if (i & (1u<<j)) sum++;
            SetField(thirdrank,thirdbits,lognpertwo*i+j,sum);    
          }              
        }         
       databits = lognpertwo;
         // convert B into unsigned[]
       data = new unsigned[n/W+1];
       for (i=0;i<n;i++) {
          if (i % W == 0) data[i/W]=0;
          data[i/W] = data[i/W] | (unsigned)B[i] << (i % W);
       }       
       //caller can delete B   
   };           
   inline unsigned rank1(unsigned i) {
      return GetField(firstrank,firstbits,i/(firstbits*firstbits));      
   };      
   inline unsigned rank2(unsigned i) {
     return GetField(secondrank,secondbits,i/firstbits);
   };
   inline unsigned rank3(unsigned i) {
      unsigned j = GetField(data,firstbits,i/firstbits);
      unsigned k = i % firstbits;
      bool isbitset = 1u & j;
      unsigned rankthree;
      if (k < databits) {
        j = (j << (W-databits)) >> (W-databits);
         rankthree = GetField(thirdrank,thirdbits,j*databits+k);
      }
      else {         
         unsigned jj = j >> databits;
         j = (j << (W-databits)) >> (W-databits);         
         rankthree = GetField(thirdrank,thirdbits,(j+1)*databits-1);
         k -= databits;
         rankthree += GetField(thirdrank,thirdbits,jj*databits+k);
      }
      return isbitset?rankthree-1:rankthree;      
   };
   inline unsigned rank(unsigned i) { // returs sum_{0<=i'<=i} B[i]
      return rank1(i)+rank2(i)+rank3(i);            
   };
   inline bool IsBitSet(unsigned i) {
      return (1u << (i % W)) & data[i/W];
   };      
   unsigned selectBasic(unsigned x) { // returns i such that x=rank(i) && rank(i-1)<x
      unsigned l=0, r=n;
      unsigned mid=(l+r)/2;      
      unsigned rankmid = rank(mid);      
      while (l+1<r) {
        if (x>rankmid) {
          l=mid;
          mid = (l+r)/2;      
        } else {
          r = mid;
          mid = (l+r)/2;    
        }     
        rankmid = rank(mid);         
      }    
      if (rankmid == x) return l;
      else return r;
   };
   inline unsigned select(unsigned x) { // returns i such that x=rank(i) && rank(i-1)<x
      // first binary search over first level rank structure
      // then binary search over second level rank structure
      // then binary search over third level rank structure
      // this is 6-7 times faster than selectBasic
      unsigned rankr;
      unsigned l=0, r=n/(firstbits*firstbits)+1;
      unsigned mid=(l+r)/2;      
      unsigned rankmid = GetField(firstrank,firstbits,mid);      
      while (l<mid) {
        if (x>rankmid)
          l = mid;
         else
          r = mid;
       mid = (l+r)/2;              
        rankmid = GetField(firstrank,firstbits,mid);         
      }    
      unsigned left;
      unsigned rankleft;
      left=l*firstbits;
      rankleft = rankmid;
          
      // now correct value is between left*firstbits+1...(left+firstbits)*firstbits
      l=0; r = firstbits;
      mid=(l+r)/2;      
      rankmid = GetField(secondrank,secondbits,left+mid);      
      while (l<mid) {
        if (x>rankleft+rankmid)
          l = mid;
        else
          r = mid;
        mid = (l+r)/2;         
        rankmid = GetField(secondrank,secondbits,left+mid);         
      } 
      left += l;
      rankleft += GetField(secondrank,secondbits,left);

      // now correct value is between left*firstbits+1...(left+1)*firstbits-1
      unsigned j = GetField(data,firstbits,left);
      unsigned jj = j >> databits; //sedondhalf
      j = (j << (W-databits)) >> (W-databits);
      rankleft -= 1u & j;      
      rankmid = GetField(thirdrank,thirdbits,(j+1)*databits-1);
      r = databits;
      if (rankleft+rankmid < x) {
         j = jj;
         left = left*firstbits+databits;
         rankleft += rankmid;
         r = firstbits-databits;
      } 
      else left = left*firstbits;      
      l=0;
      mid=(l+r)/2;      
      rankmid = GetField(thirdrank,thirdbits,j*databits+mid);      
      while (l<mid) {
        if (x>rankleft+rankmid)
          l = mid;
        else
          r = mid;
        mid = (l+r)/2;         
        rankmid = GetField(thirdrank,thirdbits,j*databits+mid);               
      }    
      if (rankleft+rankmid == x) return left+l;
      else return left+r;      
   }
   inline unsigned prev(unsigned start) {
      // returns the position of the previous 1 bit before and including start.
      // tuned to 32 bit machine
      
      unsigned i = start >> 5;
      int offset = (start % W);
      unsigned answer = start;
      unsigned int val = data[i] << (Wminusone-offset);

      if (!val) { val = data[--i]; answer -= 1+offset; }

      while (!val) { val = data[--i]; answer -= W; }

      if (!(val & 0xFFFF0000)) { val <<= 16; answer -= 16; }
      if (!(val & 0xFF000000)) { val <<= 8; answer -= 8; }

      while (!(val & 0x80000000)) { val <<= 1; answer--; }
      return answer;
      /*
      // simpler, but 3 times slower:
      while (!IsBitSet(start)) --start;      
      return start;
      */
            
   }
   void Test() {      
      unsigned i, iprev=0;
      for (i=0;i<n;i++) {
        if (rank(i)>rank(iprev)) iprev = i;
         if (rank(i)>0 && select(rank(i))!=iprev) 
            printf("select(%d)=%d!=%d\n",rank(i),select(rank(i)),iprev);                     
       }
   };   
   ~TBitRank() {
      delete firstrank;     
      delete secondrank;     
      delete thirdrank;     
      delete data;
   };      
};


/**********************************************************************************************************
 * Compressed Compact Suffix Array implementation                                                         *
 **********************************************************************************************************/
 
class TCCSA {
public:
    unsigned posbits;
    unsigned *pos;
    unsigned length; 
    int *chars;
    TBitRank *charrank;
    TBitRank *posrank;
    TBitRank *samplerank;
    unsigned *sample;    
    unsigned samplebits;
    unsigned samplerate; // suffixes 0,samplerate,2*samplerate,...., are sampled  
    unsigned samplen;
public:
    void CreatePosArray(unsigned bits, unsigned n) 
      {posbits = bits; pos = new unsigned[(n*bits)/W+1];};
    void DeletePosArray() {delete pos;};
    void SetPos(unsigned index, unsigned x)
      {SetField(pos,posbits,index,x);};
    unsigned GetPos(unsigned index)
      {return GetField(pos,posbits,index);};        
    inline unsigned GetLen(unsigned index)
      {return posrank->select(index+2)-posrank->select(index+1);}; 
    void SetSample(bool *B, unsigned n, unsigned *sample, unsigned samplebits, unsigned samplen, unsigned samplerate) {
       this->sample = sample; this->samplebits = samplebits; this->samplen = samplen;
       this->samplerate = samplerate; 
       samplerank = new TBitRank(B,n);
       samplerank->MakeRankDictionary();
       //samplerank->Test();
       // the caller can delete B
    };    
    void CreateRankForChars(char *text, unsigned int n) { 
       int i,j=0,sumlen=0,sumchars=0;
       chars = new int[256]; // chars[c+128] tells how many times character c repeats
       char *charatindex = new char[length]; // charatindex[i] tells which char is at entry i
       for (i=0;i<256; i++) chars[i]=0;
       for (i=0;i<n;i++)
          chars[text[i]+128]++; 
       for (i=0;i<length;i++) {
          sumlen += GetLen(i);    
          while (sumlen>sumchars+chars[j])
             sumchars += chars[j++];       
          charatindex[i]=j-128;
       }   
       bool *B=new bool[length];
       j=0; B[0]=1;
       chars[j++]=charatindex[0]; // chars[j] tells which is the jth distinct char
       for (i=1;i<length;i++) {
          if (charatindex[i]!=charatindex[i-1]) {
             B[i] = 1;
             chars[j++] = charatindex[i];              
          } else B[i]=0;       
       }                
       delete charatindex;
       charrank = new TBitRank(B,length);        
       charrank->MakeRankDictionary(); 
       //charrank->Test();
       delete B;      
    };   
    void CreateRankForPos(bool *B) {
       posrank = new TBitRank(B,n);
       posrank->MakeRankDictionary();      
       //posrank->Test();
       // caller can delete B
    };
    inline bool IsSampled(unsigned x,unsigned offset) {
       return samplerank->IsBitSet(posrank->select(x+1)+offset);  
    };        
    inline unsigned GetTextPos(unsigned i) {
       unsigned depth=0;
       unsigned x;
       while (!samplerank->IsBitSet(i)) {
          x = posrank->rank(i)-1;         
          i = GetPos(x)+i-posrank->prev(i); //+posrank->select(x+1);                          
          depth++;         
       }
       return GetField(sample,samplebits,samplerank->rank(i)-1)-depth; 
    }; 
    void ShowPath(unsigned i, unsigned len) {
       int x,offset,j; char c;        
       x = posrank->rank(i)-1; // CCSA position
       offset = i-posrank->prev(i); // offset = i-posrank->select(x+1);       
       for (j=0;j<len;j++) {
          c = chars[charrank->rank(x)-1]; //CharAtIndex(x) 
          if (c=='\n') printf("%c",'/');
          else printf("%c",c);
          i = GetPos(x)+offset; // suffix array position
          x = posrank->rank(i)-1; // CCSA position
          offset = i -posrank->prev(i);
       }        
       printf("\n");
    };
    inline char CharAtIndex(unsigned i) { 
       return chars[charrank->rank(i)-1];
    };    
    inline int CompareIndex(char *pattern, unsigned int m, unsigned i) { // i is suffix array position
       // compares pattern[0...m-1] with the suffix
       int x,offset,j; char c;        
       x = posrank->rank(i)-1; // CCSA position
       offset = i-posrank->prev(i); // offset = i-posrank->select(x+1);       
       for (j=0;j<m;j++) {
          c = chars[charrank->rank(x)-1]; //CharAtIndex(x) 
          if (pattern[j]>c) return j+1;
          else if (pattern[j]<c) return -(j+1); 
          /* c==pattern[j] */
          i = GetPos(x)+offset; // suffix array position
          x = posrank->rank(i)-1; // CCSA position
          offset = i -posrank->prev(i);
       }        
       return 0;  // match 
    }; 
      
    inline unsigned ReportOccurrence(unsigned i, unsigned *path, unsigned *pathdepth) {
       unsigned depth=0;
       unsigned x;
       while (*pathdepth > depth && !samplerank->IsBitSet(i) && !posrank->IsBitSet(i)) 
          i = ++(path[depth++]);
       while (!samplerank->IsBitSet(i)) {           
          x = posrank->rank(i)-1;         
          i = GetPos(x)+i-posrank->prev(i); //posrank->select(x+1);                          
          path[depth++]=i;
       }
       *pathdepth = depth;
       return GetField(sample,samplebits,samplerank->rank(i)-1)-depth;             
    }; 
       
    void ReportOccurrences(unsigned l, unsigned r) {
       unsigned *path = new unsigned[samplerate+1];
       unsigned pathdepth = 0;
       unsigned i, x;
       for (i=l;i<=r;i++) {
          x = ReportOccurrence(i,path,&pathdepth);        
          if (print) {
             printf("Match at %d",x);
             fflush(stdout);
             if (report>2) {
                printf(": ");
                ShowPath(i,m+report);
             }
             else 
                printf("\n");
          }
       }
       delete path;
     };
     int Search(char *pattern, unsigned int m) {
        int L;
        int R;
        int i;
        int Llimit;
        int Rlimit=n;
        unsigned matches=0;
        // search for left limit
        L = -1;
        R = n;
        i = (L+R)/2;
        while (L<R-1) {
           if (CompareIndex(pattern,m,i) <= 0)
              R = i;
           else
              L = i;
           i = (R+L)/2;
        }
        Llimit=R;
        // search for right limit
        L = Llimit-1;
        R = n;
        i = (L+R)/2;
        if (Llimit < n) {
           while (L < R-1) {
              if (CompareIndex(pattern,m,i) >= 0)
                 L = i;
              else
                 R = i;
              i = (R+L)/2;
           }
           Rlimit=L;
        }
        if (Llimit == n || Llimit > Rlimit) {
            if (print) printf("Pattern (%s) not found\n",pattern);
          return 0; 
        }
        matches=Rlimit-Llimit+1;
        if (!report) {
           if (print && matches) 
              printf("Pattern (%s) found\n",pattern);              
           else if (!matches && print)
              printf("Pattern (%s) not found\n",pattern);
           return matches>0;
        }
        if (report == 2) return matches; // counting        
        ReportOccurrences(Llimit,Rlimit);
        // above is 25% faster than this:
        /*
        for (L=Llimit;L<=Rlimit;L++) {
           i = GetTextPos(L);
           if (print) {
              printf("Match at %d",i);
              if (report>2) {
                 printf(": ");  
                 ShowPath(L,m+report); // report > 2 == how large context is shown
              }
              else 
                 printf("\n");             
           }
        } 
        */
        return matches;
     };
     void Print(char *text) {
        int i;
        printf("Compressed Compact Suffix Array of text \n %s \n is:\n",text);
        fflush(stdout);
        for (i=0;i<length;i++) {
           printf("%d: (%d,%d) - %c\n",i,GetPos(i),GetLen(i),CharAtIndex(i)); 
           fflush(stdout);
        }
     };     
      
     unsigned SpaceRequirementInBits() {
        return charrank->SpaceRequirementInBits()+ //charrank
                256*W+ // chars
                posrank->SpaceRequirementInBits()+ // posrank
                samplebits*samplen+samplerank->SpaceRequirementInBits()+ // samplerank
                posbits*length; // the array itself
     };
      
     ~TCCSA() {
        delete pos; delete sample; delete chars; 
        delete charrank; delete posrank; delete samplerank;
     };

};

unsigned MarkLinksToSuffixArray(TSA *SArray, TSA *DArray, char *text, unsigned n)
/* This is for making compressed compact suffix arrays */
{
   unsigned newsize=n;
   unsigned i;
   unsigned j;
   unsigned x;
   unsigned index;
   // Make DArray the inverse of SArray
   for (i=0; i<n; i++) {
      DArray->SetPos(SArray->GetPos(i),i);
      DArray->SetB(i,0);
      SArray->SetB(i,0);
   }
   // Swap SArray and DArray
   for (i=0; i<n; i++) {
      x = DArray->GetPos(i);
      DArray->SetPos(i,SArray->GetPos(i));
      SArray->SetPos(i,x);
   }
   i=0;
   /* Mark links */
   while (i<n) {
      x = SArray->GetPos((DArray->GetPos(i)+1) % n);
      DArray->SetB(i,0);
      unsigned textpos=DArray->GetPos(i);
      DArray->SetPos(i,x);
      j = 1;
      while (x+j <n && i+j <n &&                
             SArray->GetPos(((DArray->GetPos(i+j)+1) % n)) == x+j && 
             text[DArray->GetPos(i+j)] == text[textpos]) { // restrict a link to contain only one character            
         DArray->SetB(i+j,1);
         DArray->SetPos(i+j,x+j);
         newsize--;
         j++;
      }
      i = i+j;
   }
   // SArray := inverse of SArray
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   } 
   return newsize;
}

void FromMarkedSAToCCSA(TSA *SArray, TCCSA *DArray)
{
   unsigned i=0;
   unsigned j=0;
   bool *B = new bool[n];
   while (i<n) {
      DArray->SetPos(j++,SArray->GetPos(i));
      B[i++]=1;
      while (i<n && SArray->GetB(i))         
        B[i++]=0;    
   }
   B[i]=1;
   DArray->CreateRankForPos(B);
   delete B;
}


TCCSA *MakeCompressedCompactSuffixArray(TSA *SA, char *text, unsigned n)
{
   unsigned newsize=n,i;
   unsigned samplerate=bitsneeded(n);
   unsigned samplebits=samplerate;
   unsigned *sample = new unsigned[n/W+2];
   unsigned samplen = n/samplerate+1;
   bool *B = new bool[n];
   unsigned rank=0;
   /* Make sample of suffixes for reporting */
   for (i=0;i<n;i++)
      if (SA->GetPos(i) % samplerate == 0 || SA->GetPos(i) == n-1) {
        SetField(sample,samplebits,rank,SA->GetPos(i));
        B[i]=1;            
        rank++;
      } else B[i]=0;       
   TSA *SATemp = new TSA();
   SATemp->CreatePosArray(bitsn,n);   
   newsize = MarkLinksToSuffixArray(SA, SATemp, text, n); 
   delete SA;
   TCCSA *CCSA = new TCCSA();
   CCSA->length = newsize;
   CCSA->CreatePosArray(bitsn,newsize+1);
   FromMarkedSAToCCSA(SATemp,CCSA);
   delete SATemp;
   CCSA->CreateRankForChars(text,n);
   CCSA->SetSample(B,n,sample,samplebits,samplen,samplerate);
   delete B;
   // not needing text anymore
   //delete text;
   printf("Compressed Compacted Suffix Array is of length %d \n and takes %f B space \n",
          CCSA->length,CCSA->SpaceRequirementInBits()/8.0);
   fflush(stdout);
   return CCSA;
}

TCCSA *MakeSuffixArrayAndCompressIt(char *text, unsigned n)
{
   TSA *SA = CreateSuffixArray(text,n);
   printf("Suffix Array takes %d*%f B = %f B space \n",
          n,(1.0*SA->posbits)/8,
          n*((1.0*SA->posbits)/8));
   printf("Text size is %d B\n",n);
   return MakeCompressedCompactSuffixArray(SA,text,n);    
}

/**********************************************************************************************************
 * FM-index implementation using alphabet ranking                                                         *
 **********************************************************************************************************/

class TAlphabetRank {
// using fixed 0...255 alphabet
public: 
   int min; // contains ranks for symbols min...max
   int max;
   bool *B;
   TBitRank *bitrank;
   TAlphabetRank *left;
   TAlphabetRank *right;
public:
   TAlphabetRank(char *s, unsigned n, int minp, int maxp) {
      left = NULL;
      right = NULL;
      bitrank = NULL;
      B = new bool[n];
      min = minp;
      max = maxp;
      int mid = (min+max)/2;
      unsigned i, sum=0;
      for (i=0;i<n;i++) {   
         B[i] = ((int)s[i]+128)>mid;
         sum += B[i];
      }
      char *sfirst, *ssecond;
      if (n-sum > 0)  
         sfirst = new char[n-sum];
      if (sum > 0) 
         ssecond = new char[sum];
      unsigned j=0,k=0;
      for (i=0;i<n;i++)
         if (B[i]) ssecond[k++] = s[i];
         else sfirst[j++] = s[i];
      bitrank = new TBitRank(B,n);
      bitrank->MakeRankDictionary();
      //bitrank->Test();
      delete B;
      if (j > 0) {
         if (min < max) 
            left = new TAlphabetRank(sfirst,j,min,mid); 
         delete sfirst;
      }
      if (k>0) { 
         if (min < max)           
            right = new TAlphabetRank(ssecond,k,mid+1,max); 
         delete ssecond;
      }
   };   
   inline unsigned rank(int c, int i) { // returns the number of characters c before and including position i
      TAlphabetRank *temp=this;
      if (c < temp->min || c >temp->max) return 0;
      while (temp!=NULL && i>=0) {
         if (c<=(temp->min+temp->max)/2) {
            i = i-temp->bitrank->rank(i); 
            temp = temp->left; 
         }
         else { 
            i = temp->bitrank->rank(i)-1; 
            temp = temp->right;
         }
      } 
      return i+1;
   };   
   unsigned SpaceRequirementInBits() {
      unsigned bits=0;
      if (left!=NULL) bits+=left->SpaceRequirementInBits();
      if (right!=NULL) bits+=right->SpaceRequirementInBits();
      bits += bitrank->SpaceRequirementInBits();
      return bits;
   };
   bool Test(char *s, unsigned n) {
      // testing that the code works correctly
      int C[256];
      int i,j;
      bool correct=true;
      for (j=0;j<256;j++)
         C[j] = 0;
      for (i=0;i<n;i++) {
         C[(int)s[i]+128]++;
         if (C[(int)s[i]+128]!=rank((int)s[i]+128,i)) {
            correct = false;
            printf("%d (%c): %d<>%d\n",i,s[i],C[(int)s[i]+128],rank((int)s[i]+128,i)); 
         }         
      }
      return correct;            
   };   
   ~TAlphabetRank() {
      if (left!=NULL) delete left;
      if (right!=NULL) delete right;
      if (bitrank!=NULL)    
         delete bitrank;
   };   

};

void TestAlphabetRank() {
   char *s="acccbbbcccccccccaacaabaacaaaaaaacbbba";
   TAlphabetRank *ar = new TAlphabetRank(s,strlen(s),0,255);
   ar->Test(s,strlen(s));
   unsigned correct[256];
   unsigned i;
   for (i=0;i<255;i++)
      correct[i]=0;
   printf("%s\n",s);
   printf("a=%d\n",'a');
   printf("b=%d\n",'b');   
   for (i=0;i<strlen(s);i++) {
      correct[(int)s[i]+128]++;    
      printf("rank('a',%d)=%d (%d)\n",i,ar->rank((int)'a'+128,i), correct[(int)'a'+128]);
      printf("rank('b',%d)=%d (%d)\n",i,ar->rank((int)'b'+128,i), correct[(int)'b'+128]);
      printf("rank('c',%d)=%d (%d)\n",i,ar->rank((int)'c'+128,i), correct[(int)'c'+128]);      
   }
}   

char *BWT(char *text, unsigned n) {
   TSA *SA = CreateSuffixArray(text,n);
   char *s = new char[n];
   unsigned i;
   for (i=0;i<n;i++)
      if (SA->GetPos(i)==0) s[i] = text[n-1];
      else s[i] = text[SA->GetPos(i)-1];
   delete SA;
   return s;      
}

class TFMindex {
public:
    unsigned n;
    unsigned C[256];
    TAlphabetRank *alphabetrank;
public:
    TFMindex(char *text, unsigned n) {
       printf("Text size is %d B\n",n);        
       this->n = n;
       char *bwt = BWT(text,n);
       // caller can delete text now
       unsigned i, min, max;
       for (i=0;i<256;i++)
          C[i]=0;
       for (i=0;i<n;++i)
          C[(int)bwt[i]+128]++;
       for (i=0;i<256;i++)
          if (C[i]>0) {min = i; break;}          
       for (i=255;i>=min;--i)
          if (C[i]>0) {max = i; break;}                    
       unsigned prev=C[0], temp;
       C[0]=0;
       for (i=1;i<256;i++) {          
          temp = C[i];
          C[i]=C[i-1]+prev;
          prev = temp;
       }
       alphabetrank = new TAlphabetRank(bwt,n,min,max);   
       //if (alphabetrank->Test(bwt,n)) printf("alphabetrank ok\n");    
       delete bwt;
       printf("FMindex takes %d B\n",256*W/8+alphabetrank->SpaceRequirementInBits()/8);
       fflush(stdout);       
    };   
    int Search(char *pattern, unsigned m) {
       // use the FM-search to replacing function Occ(c,1,i) with alphabetrank->rank(c,i)
       int c = (int)pattern[m-1]+128; unsigned i=m-1;
       int sp = C[c];
       int ep = C[c+1]-1;
       while (sp<=ep && i>=1) {
          c = (int)pattern[--i]+128;
          sp = C[c]+alphabetrank->rank(c,sp-1);
          ep = C[c]+alphabetrank->rank(c,ep)-1;
       }
       if (sp<=ep) {
          if (print) printf("%d matches found for \"%s\"\n",ep-sp+1, pattern);
          return ep-sp+1;
       } 
       else {
          if (print) printf("Pattern not found\n");
          return 0;
       }
    };   
    ~TFMindex() {
       delete alphabetrank;       
    };
};
